#include <string.h>
#include <stdio.h>
#include <math.h>

#include "sdkconfig.h"
#include "esp_log.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/semphr.h"
#include "beatz_adc.h"
#include "esp_adc/adc_continuous.h"
#include "esp_dsp.h"
#include "beatz_pwm.h"

#include "freertos/ringbuf.h"

// ADC
#define N 128 // Number of samples
#define FRAME_SIZE N * SOC_ADC_DIGI_RESULT_BYTES
#define SAMPLE_FREQ 20000
#define GET_UNIT(x)        ((x>>3) & 0x1)
#define UNIT2(x)           ((x | 1 << 3))

// GPIO
#define DBG_PIN GPIO_NUM_9
#define GPIO_OUTPUT_PIN_SEL (1ULL << DBG_PIN)

TaskHandle_t control_adc_task_h;
TaskHandle_t eval_adc_task_h;

static adc_channel_t channel[CH_NUM] = { UNIT2(ADC_CHANNEL_0), 
                                    UNIT2(ADC_CHANNEL_1), 
                                    UNIT2(ADC_CHANNEL_2), 
                                    UNIT2(ADC_CHANNEL_3), 
                                    UNIT2(ADC_CHANNEL_4), 
                                    UNIT2(ADC_CHANNEL_5), 
                                    UNIT2(ADC_CHANNEL_6), 
                                    UNIT2(ADC_CHANNEL_7)};
static adc_digi_pattern_config_t adc_patterns[CH_NUM][SOC_ADC_PATT_LEN_MAX] = {0};
static uint8_t adc_buf[FRAME_SIZE];
static const char *TAG = "ADC";

__attribute__((aligned(16))) float fft_data_cf[N*2];
__attribute__((aligned(16))) float wind[N];

const int pwm_freq_idx = (int) ((float)PWM_FREQ / SAMPLE_FREQ * FRAME_SIZE + 2);    // We add 2 because the signal is shown at index 26.5 in the fft plot. So we take index 27.


uint32_t start_cnt;
uint32_t end_cnt;

// Function definitions
static bool IRAM_ATTR s_conv_done_cb(adc_continuous_handle_t handle, const adc_continuous_evt_data_t *edata, void *user_data)
{
    BaseType_t mustYield = pdFALSE;
    //Notify that ADC continuous driver has done enough number of conversions
    configASSERT(control_adc_task_h);
    vTaskNotifyGiveFromISR(control_adc_task_h, &mustYield);

    return (mustYield == pdTRUE);
}

static void adc_configure(adc_continuous_handle_t handle, uint8_t pattern_idx){
    adc_continuous_config_t dig_cfg = {
        .sample_freq_hz = SAMPLE_FREQ,
        .conv_mode = ADC_CONV_SINGLE_UNIT_2,
        .format = ADC_DIGI_OUTPUT_FORMAT_TYPE1,
        .pattern_num = 1,
        .adc_pattern = adc_patterns[pattern_idx],
    };

    ESP_ERROR_CHECK(adc_continuous_config(handle, &dig_cfg));
}

void adc_initialize(adc_continuous_handle_t *out_handle)
{
    adc_continuous_handle_t handle = NULL;

    adc_continuous_handle_cfg_t adc_config = {
        .max_store_buf_size = FRAME_SIZE,
        .conv_frame_size = FRAME_SIZE,
    };
    ESP_ERROR_CHECK(adc_continuous_new_handle(&adc_config, &handle));

    // Initialize List of ADC patterns
    for (int i = 0; i < CH_NUM; i++){
        // Each Pattern contains just one channel. This is the simplest solution.
        adc_patterns[i][0].atten = ADC_ATTEN_DB_0;
        adc_patterns[i][0].channel = channel[i] & 0x7;
        adc_patterns[i][0].unit = GET_UNIT(channel[i]);
        adc_patterns[i][0].bit_width = SOC_ADC_DIGI_MAX_BITWIDTH;
    }

    adc_configure(handle, 0);

    // Register callback when frame is full
    adc_continuous_evt_cbs_t cbs = {
        .on_conv_done = s_conv_done_cb,
    };
    ESP_ERROR_CHECK(adc_continuous_register_event_callbacks(handle, &cbs, NULL));

    *out_handle = handle;
}

void print_data(float *buf, int len)
{
    for (int n = 0; n < len; n++){
        printf("%f\n", buf[n]);
    }
}

void vEvaluateADC (void *pvParameters)
{
    // This task evaluates the ADC conversion frames
    ESP_LOGW(TAG, "Start ADC-Eval task.");
    esp_err_t ret;
    ESP_LOGI(TAG, "Init fft");
    ret = dsps_fft2r_init_fc32(NULL, CONFIG_DSP_MAX_FFT_SIZE);
    if (ret  != ESP_OK)
    {
        ESP_LOGE(TAG, "Not possible to initialize FFT. Error = %i", ret);
        return;
    }
    // Generate Hann window
    dsps_wind_hann_f32(wind, N);

    for(;;){
        ulTaskNotifyTake(pdTRUE, (TickType_t) portMAX_DELAY);
        uint16_t channel = ((adc_digi_output_data_t *) adc_buf)->type1.channel;
        uint16_t data_idx = 0;
        for(int i = 0; i < FRAME_SIZE; i+= SOC_ADC_DIGI_RESULT_BYTES){
            adc_digi_output_data_t *p = (void*)&adc_buf[i];
            fft_data_cf[data_idx + 0] = ((float) p->type1.data) / 4095.0;   // Normalize the values to be between 0.0 and 1.0
            //fft_data_cf[data_idx + 0] = i % 6; // For testing
            fft_data_cf[data_idx + 1] = 0; // Imaginary part of the signal is 0
            data_idx+=2;    // We could use i instead of data_idx because both get incremented by two but this is safer in case SOC_ADC_DIGI_RESULT_BYTES changes.
        }
        //ESP_LOGW(TAG, "Channel: %d", channel);

        // ESP_LOGW(TAG, "Raw values:"); 
        // print_data(fft_data_cf, N);

        // Run FFT (Use ansi function because I couldn't find an optimized version for the esp32s2)
        dsps_fft2r_fc32(fft_data_cf, N);    

        // Bit reverse
        dsps_bit_rev_fc32(fft_data_cf, N);
        
        // Convert complex vector to two real vectors 
        dsps_cplx2reC_fc32(fft_data_cf, N);     

        // ESP_LOGW(TAG, "complex to real"); 
        // print_data(fft_data_cf, N);
        //ESP_LOGW(TAG, "abs(fft_data_cf[%d]) =  %f", pwm_freq_idx, fabs(fft_data_cf[pwm_freq_idx]));

        vTaskDelay(1);
    }
}

static void adc_measure(adc_continuous_handle_t handle, int idx)
{
    static uint32_t act_len;
    esp_err_t ret;
    
    adc_configure(handle, idx);   // reconfigure the ADC (Change the pattern/channel)
    ESP_ERROR_CHECK(adc_continuous_start(handle));  // Start the ADC

    ulTaskNotifyTake(pdTRUE, portMAX_DELAY);
    ret = adc_continuous_read(handle, adc_buf, FRAME_SIZE, &act_len, 0);  // Read out the frame buffer
    ESP_ERROR_CHECK(adc_continuous_stop(handle));   // Stop the ADC
    if (ret == ESP_OK && act_len == FRAME_SIZE) {
        xTaskNotifyGive(eval_adc_task_h);
    }
}

void vControlADC (void *pvParameters)
{
    // This task is for controlling the ADC.
    ESP_LOGW(TAG, "Start ADC-Control task.");
    adc_continuous_handle_t handle = NULL;
    adc_initialize(&handle);

    pwm_initialize();
    pwm_start();

    for(;;){
        for(int c_idx = 0; c_idx < CH_NUM; c_idx++){    // c_idx = cycle index. In each cycle, another pwm pin is configured
            pwm_configure(c_idx);
            for(int m_idx = c_idx; m_idx < CH_NUM; m_idx++){    // m_idx = measurement index. In each measurement another adc channel is measured
                adc_measure(handle, m_idx);
                vTaskDelay(pdMS_TO_TICKS(3000));    // To give the evaluation task some time because printing is slow (Just for testing)
            }
        }
    }

}
